/*
 *  Galaxy Script Obfuscation Tool
 *  Copyright (C) 2011 Alexis Naveros
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>


#include "rand.h"


/*
#define SC_DEBUG_STDOUT
*/


/* Default settings to override with command line arguments */
#define SC_NAME_ALPHABET "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
#define SC_NAME_PREFIX "qq"
#define SC_FILE_IN "MapScript.galaxy"
#define SC_FILE_OUT "MapScript.galaxy"
#define SC_FILE_HEADER "scriptheader.txt"
#define SC_LINE_LENGTH_MAX (400)

#define SC_SHUFFLE_COUNT_FACTOR (16)
#define SC_SHUFFLE_COUNT_MAXIMUM (8*1024*1024)


#define SC_NAME_KEY_ALLOC (1048576)


#define SC_FD_LOG_FILE "log.txt"

FILE *scfdlog;


const char *scNameDeclList[] =
{
  "abilcmd",
  "actor",
  "actorscope",
  "aifilter",
  "animtarget",
  "bank",
  "bool",
  "byte",
  "camerainfo",
  "char",
  "color",
  "doodad",
  "fixed",
  "handle",
  "int",
  "marker",
  "order",
  "playergroup",
  "point",
  "region",
  "revealer",
  "soundlink",
  "string",
  "text",
  "timer",
  "transmissionsource",
  "trigger",
  "unit",
  "unitfilter",
  "unitgroup",
  "unitref",
  "void",
  "wave",
  "waveinfo",
  "wavetarget",
  0,
};

const char *scFixedNameList[] =
{
  "InitMap",
  "DebugCallback",
  "AIMeleeStart",
  "AINewUnit",
  "AIIsAttacking",
  "AINeedsDefending",
  "AICombatPriority",
  "AITownIsFull",
  "AIWaveThink",
  "AITownWasLost",
  "AIMelee",
  "AIWaveAddedUnit",
  "AIGetMaker",
  "AIGetObjectType",
  "AIGetFirstUnfinishedReq",
  "AIGetFirstMissingReq",
  "AIGetDefaultBuildFlags",
  "AISpellPriority",
  "AIGetScout",
  "AIWaveChange",
  "AIEarlyDefScout",
  0
};


////


int scWordFirstChar( char c )
{
  if( ( c >= 'a' ) && ( c <= 'z' ) )
    return 1;
  if( ( c >= 'A' ) && ( c <= 'Z' ) )
    return 1;
  if( c == '_' )
    return 1;
  return 0;
}

int scWordChar( char c )
{
  if( ( c >= 'a' ) && ( c <= 'z' ) )
    return 1;
  if( ( c >= 'A' ) && ( c <= 'Z' ) )
    return 1;
  if( ( c >= '0' ) && ( c <= '9' ) )
    return 1;
  if( c == '_' )
    return 1;
  return 0;
}


#define SC_CHAR_IS_DELIMITER(c) ((c)<=' ')

char *scStrNextWord( char *str )
{
  for( ; ; str++ )
  {
    if( *str == 0 )
      return 0;
    if( !( SC_CHAR_IS_DELIMITER( *str ) ) )
      break;
  }
  return str;
}

char *scStrCmpSeq( char *str, char *seq )
{
  int i;
  for( i = 0 ; ; i++ )
  {
    if( !( seq[i] ) )
      return &str[i];
    if( str[i] != seq[i] )
      break;
  }
  return 0;
}

char *scWordCmpSeq( char *str, int len, char *seq )
{
  int i;
  for( i = 0 ; ; i++ )
  {
    if( i >= len )
    {
      if( !( seq[i] ) )
        return &str[i];
      break;
    }
    if( str[i] != seq[i] )
      break;
  }
  return 0;
}

int scStrWordLen( char *str )
{
  int i;
  if( !( scWordFirstChar( str[0] ) ) )
    return 0;
  for( i = 1 ; ; i++ )
  {
    if( !( scWordChar( str[i] ) ) )
      break;
  }
  return i;
}

char *scStrNextLine( char *str )
{
  for( ; ; str++ )
  {
    if( *str == 0 )
      return str;
    if( *str == '\n' )
      break;
  }
  return str + 1;
}

////



char *scNameDeclCheck( char *str, int len )
{
  int index;
  char *decl, *end;
  for( index = 0 ; scNameDeclList[index] ; index++ )
  {
    decl = (char *)scNameDeclList[index];
    end = scWordCmpSeq( str, len, decl );
    if( end )
      return end;
  }
  return 0;
}


/* Characters that count as separation between words, [] must be excluded for array definitions, int[3] foo; */
const int scNameSeparationChar[0x100] =
{
  ['='] = 1,
  [';'] = 1,
  ['{'] = 1,
  ['}'] = 1,
  ['('] = 1,
  [')'] = 1,
  [','] = 1,
  ['+'] = 1,
  ['-'] = 1,
  ['*'] = 1,
  ['/'] = 1,
  ['%'] = 1,
  ['|'] = 1,
  ['&'] = 1,
  ['^'] = 1,
  ['<'] = 1,
  ['>'] = 1
};

/* Characters after which it is perfectly safe to insert a line break */
const int scSafeBreakChar[0x100] =
{
  [';'] = 1,
  ['{'] = 1,
  ['}'] = 1,
  ['('] = 1,
  [')'] = 1,
  [','] = 1
};



////



typedef struct
{
  char *name;
  char *code;
} scKey;

typedef struct
{
  char *prefix;
  int prefixlen;
  char *alphabet;
  int alphabetsize;
  scKey *table;
  int count;
  int alloc;
  int codelen;
  int *swaptable;
  /* Key hash table? Meh, stupid search is fast enough. */
} scKeyDict;


void scKeyDictInit( scKeyDict *keydict, char *alphabet, char *prefix )
{
  keydict->table = malloc( SC_NAME_KEY_ALLOC * sizeof(scKey) );
  keydict->count = 0;
  keydict->alloc = SC_NAME_KEY_ALLOC;
  keydict->alphabet = alphabet;
  keydict->alphabetsize = strlen( alphabet );
  keydict->prefix = prefix;
  keydict->prefixlen = strlen( prefix );
  return;
}

scKey *scNameKey( scKeyDict *keydict, char *word, int wordlen )
{
  int keyindex;
  scKey *key;
  key = keydict->table;
  for( keyindex = 0 ; keyindex < keydict->count ; keyindex++, key++ )
  {
    if( scWordCmpSeq( word, wordlen, key->name ) )
      return key;
  }
  return 0;
}

void scRegisterKey( scKeyDict *keydict, char *name, int namelen )
{
  scKey *key;
  if( scNameKey( keydict, name, namelen ) )
    return;
  key = &keydict->table[keydict->count];
  key->code = 0;
  key->name = malloc( namelen+1 );
  memcpy( key->name, name, namelen );
  key->name[namelen] = 0;
  keydict->count++;
  if( keydict->count >= keydict->alloc )
  {
    keydict->alloc <<= 1;
    keydict->table = realloc( keydict->table, keydict->alloc );
  }
  return;
}

void scRegisterFixedKey( scKeyDict *keydict, char *name, char *code )
{
  scKey *key;
  key = &keydict->table[keydict->count];
  key->name = name;
  key->code = code;
  keydict->count++;
  if( keydict->count >= keydict->alloc )
  {
    keydict->alloc <<= 1;
    keydict->table = realloc( keydict->table, keydict->alloc );
  }
  return;
}

int scKeyIndexSwap( int srcindex, int *swaptable )
{
  int i, dstindex;
  dstindex = 0;
  for( i = 0 ; i < 32 ; i++ )
    dstindex |= ( ( srcindex >> swaptable[i] ) & 0x1 ) << i;
  return dstindex;
}

void scKeyDictBuild( scKeyDict *keydict, int mincodelen )
{
  int i, j, domainsize, codebits, keyindex, codeindex, swapcount;
  int swap0, swap1;
  char *code;
  scKey *key;
  rand32State randstate;

  /* Clamp mincodelen! */
  domainsize = 1;
  for( i = 0 ; i < mincodelen ; i++ )
  {
    domainsize *= keydict->alphabetsize;
    if( domainsize >= 4*1024*1024 )
      break;
  }
  mincodelen = i;
  /* Clamp mincodelen! */

  domainsize = 1;
  for( keydict->codelen = 0 ; domainsize < keydict->count ; keydict->codelen++ )
    domainsize *= keydict->alphabetsize;

  /* Minimum character count */
  for( ; keydict->codelen < mincodelen ; keydict->codelen++ )
    domainsize *= keydict->alphabetsize;

  rand32Seed( &randstate, time(0) );
  keydict->swaptable = malloc( domainsize * sizeof(int) );
  for( keyindex = 0 ; keyindex < domainsize ; keyindex++ )
    keydict->swaptable[keyindex] = keyindex;
  swapcount = SC_SHUFFLE_COUNT_FACTOR * keydict->count;
  if( swapcount > SC_SHUFFLE_COUNT_MAXIMUM )
    swapcount = SC_SHUFFLE_COUNT_MAXIMUM;
  for( i = 0 ; i < swapcount ; i++ )
  {
    swap0 = rand32Int( &randstate ) % keydict->count;
    swap1 = rand32Int( &randstate ) % domainsize;
    j = keydict->swaptable[swap0];
    keydict->swaptable[swap0] = keydict->swaptable[swap1];
    keydict->swaptable[swap1] = j;
  }

  key = keydict->table;
  for( keyindex = 0 ; keyindex < keydict->count ; keyindex++, key++ )
  {
    if( key->code )
      continue;
    codeindex = keydict->swaptable[ keyindex ];
    j = codeindex;
    key->code = malloc( keydict->prefixlen + keydict->codelen + 1 );
    code = key->code;
    memcpy( code, keydict->prefix, keydict->prefixlen );
    code += keydict->prefixlen;
    for( i = 0 ; i < keydict->codelen ; i++ )
    {
      code[i] = keydict->alphabet[ j % keydict->alphabetsize ];
      j /= keydict->alphabetsize;
    }
    code[i] = 0;

#ifdef SC_DEBUG_STDOUT
    fprintf( scfdlog, "Key %d -> %d ; %s -> %s\n", keyindex, codeindex, key->name, key->code );
    if( j )
    {
      fprintf( scfdlog, "INTERNAL ERROR: Remainder %d\n", codeindex );
      exit( 1 );
    }
#endif
  }

  free( keydict->swaptable );
  keydict->swaptable = 0;

  return;
}




////


int scValidateConfig( unsigned char *prefix, unsigned char *alphabet )
{
  int i, alphabetlen, prefixlen;
  char table[0x100];

  alphabetlen = strlen( alphabet );
  if( alphabetlen < 2 )
  {
    fprintf( scfdlog, "ERROR: Alphabet must count at least 2 characters\n" );
    return 0;
  }

  memset( table, 0, 0x100*sizeof(char) );
  prefixlen = strlen( prefix );
  if( prefixlen )
  {
    if( !( scWordFirstChar( prefix[0] ) ) )
    {
      fprintf( scfdlog, "ERROR: Invalid first character for prefix : %c\n", prefix[0] );
      return 0;
    }
    for( i = 0 ; i < prefixlen ; i++ )
    {
      if( scWordChar( prefix[i] ) )
        continue;
      fprintf( scfdlog, "ERROR: Invalid character for prefix : %c\n", prefix[i] );
      return 0;
    }
    for( i = 0 ; i < alphabetlen ; i++ )
    {
      if( !( scWordChar( alphabet[i] ) ) )
      {
        fprintf( scfdlog, "ERROR: Invalid character for alphabet : %c\n", alphabet[i] );
        return 0;
      }
      if( table[ (int)alphabet[i] ] )
      {
        fprintf( scfdlog, "ERROR: Can't repeat character in alphabet : %c\n", alphabet[i] );
        return 0;
      }
      table[ (int)alphabet[i] ] = 1;
    }
  }
  else
  {
    for( i = 0 ; i < alphabetlen ; i++ )
    {
      if( !( scWordFirstChar( alphabet[i] ) ) )
      {
        if( !( scWordChar( alphabet[i] ) ) )
          fprintf( scfdlog, "ERROR: Invalid character for alphabet : %c\n", alphabet[i] );
        else
          fprintf( scfdlog, "ERROR: Can't use numerical characters in alphabet without using a prefix : %c\n", alphabet[i] );
        return 0;
      }
      if( table[ (int)alphabet[i] ] )
      {
        fprintf( scfdlog, "ERROR: Can't repeat character in alphabet : %c\n", alphabet[i] );
        return 0;
      }
      table[ (int)alphabet[i] ] = 1;
    }
  }

  return 1;
}


////


/* Not quite safe, in case someone encodes names with a really tiny alphabet and it grows... a lot */
#define SC_MALLOC_SIZE(x) (4*x+1048576)


int main( int argc, char *argv[] )
{
  int i, srcsize, dstindex, wordlen, wordflag, newlineflag, nameflag, spaceflag, codelen, linelength, linemaxlength, keymatchcount, stringflag;
  int mincodelen, headersize;
  FILE *fd;
  char *filein, *fileout;
  char *alphabet, *prefix;
  char *header;
  char *buf0, *buf1;
  char *src, *varsrc, *wordsrc, *linesrc, *dst;
  scKey *key;
  scKeyDict keydict;

  scfdlog = fopen( SC_FD_LOG_FILE, "w" );
  if( !( scfdlog ) )
  {
    printf( "Failed to open log file for writing : %s\n", SC_FD_LOG_FILE );
    return 0;
  }

  /* sc2galaxystrip.exe MapScript.galaxy "prefix" "abcdef" 4 200 */
  filein = SC_FILE_IN;
  fileout = SC_FILE_OUT;
  if( argc >= 2 )
  {
    filein = argv[1];
    fileout = argv[1];
    fprintf( scfdlog, "Config : input file set to \"%s\"\n", argv[1] );
    fprintf( scfdlog, "Config : output file set to \"%s\"\n", argv[1] );
  }
  prefix = SC_NAME_PREFIX;
  if( argc >= 3 )
  {
    prefix = argv[2];
    fprintf( scfdlog, "Config : prefix set to \"%s\"\n", argv[2] );
  }
  alphabet = SC_NAME_ALPHABET;
  if( argc >= 4 )
  {
    alphabet = argv[3];
    fprintf( scfdlog, "Config : alphabet set to \"%s\"\n", argv[3] );
  }
  mincodelen = 3;
  if( argc >= 5 )
  {
    mincodelen = atoi( argv[4] );
    fprintf( scfdlog, "Config : code miminum length set to %d\n", mincodelen );
  }
  linemaxlength = SC_LINE_LENGTH_MAX;
  if( argc >= 6 )
  {
    linemaxlength = atoi( argv[5] );
    fprintf( scfdlog, "Config : line length set to %d\n", linemaxlength );
  }


  /* Sanity checks on parameters */
  if( !( scValidateConfig( (unsigned char *)prefix, (unsigned char *)alphabet ) ) )
    return 0;

  fprintf( scfdlog, "Reading input file %s\n", filein );
  fd = fopen( filein, "rb" );
  if( !( fd ) )
  {
    fprintf( scfdlog, "ERROR: Failed to read : %s\n", filein );
    return 0;
  }
  fseek( fd, 0, SEEK_END );
  srcsize = ftell( fd );
  fseek( fd, 0, SEEK_SET );
  buf0 = malloc( SC_MALLOC_SIZE(srcsize) );
  fread( buf0, 1, srcsize, fd );
  fclose( fd );
  buf0[srcsize] = 0;
  buf1 = malloc( SC_MALLOC_SIZE(srcsize) );
  fprintf( scfdlog, "  Input Size : %d bytes\n", srcsize );
  

  /* Begin first pass */
  fprintf( scfdlog, "Parsing and condensing script\n" );
  scKeyDictInit( &keydict, alphabet, prefix );
  for( i = 0 ; scFixedNameList[i] ; i++ )
    scRegisterFixedKey( &keydict, (char *)scFixedNameList[i], (char *)scFixedNameList[i] );
  src = buf0;
  dst = buf1;
  dstindex = 0;
  wordflag = 0;
  newlineflag = 0;
  nameflag = 0;
  spaceflag = 0;
  for( ; ; )
  {
    wordsrc = scStrNextWord( src );
    if( !( wordsrc ) )
      break;
    spaceflag = ( wordsrc != src );
    src = wordsrc;

    if( scStrCmpSeq( src, "//" ) ) 
    {
      src = scStrNextLine( src );
      continue;
    }
    else if( *src == '\n' ) 
    {
      if( newlineflag )
      {
        dst[dstindex++] = '\r';
        dst[dstindex++] = '\n';
        newlineflag = 0;
      }
      src++;
      continue;
    }
    else if( *src == '\"' ) 
    {
      for( ; ; )
      {
        dst[dstindex++] = *src++;
        if( !( *src ) )
          goto end0;
        if( ( *src == '\"' ) && ( src[-1] != '\\' ) )
          break;
      }
      dst[dstindex++] = *src++;
      continue;
    }

    wordlen = scStrWordLen( src );
    if( wordlen )
    {
      if( scStrCmpSeq( src, "include" ) ) 
      {
        linesrc = scStrNextLine( src );
        wordlen = linesrc - src;
        memcpy( &dst[dstindex], src, wordlen );
        src += wordlen;
        dstindex += wordlen;
        src = linesrc;
        continue;
      }
      varsrc = scNameDeclCheck( src, wordlen );
      if( varsrc )
        nameflag = 1;
      else if( nameflag )
      {
        scRegisterKey( &keydict, src, wordlen );
#ifdef SC_DEBUG_STDOUT
        fprintf( scfdlog, "Name : %.*s\n", wordlen, src );
#endif
        nameflag = 0;
      }
      if( ( wordflag ) && ( spaceflag ) )
        dst[dstindex++] = ' ';
      memcpy( &dst[dstindex], src, wordlen );
      src += wordlen;
      dstindex += wordlen;
      wordflag = 1;
    }
    else
    {
      if( scNameSeparationChar[(int)(*src)] )
        wordflag = 0;
      dst[dstindex++] = *src++;
    }
  }
  end0:
  dst[dstindex] = 0;
  /*  End first pass */
  scKeyDictBuild( &keydict, mincodelen );
  fprintf( scfdlog, "  Name Count : %d\n", keydict.count );
  fprintf( scfdlog, "  Alphabet Size : %d\n", keydict.alphabetsize );
  fprintf( scfdlog, "  Key Length : %d\n", keydict.codelen );


#ifdef SC_DEBUG_STDOUT
  if( ( fd = fopen( "TEMP.txt", "wb" ) ) )
  {
    fwrite( buf1, 1, dstindex, fd );
    fclose( fd );
  }
#endif


  /* Second pass */
  fprintf( scfdlog, "Packing script with key replacements\n" );
  src = buf1;
  dst = buf0;
  dstindex = 0;
  linelength = 0;
  keymatchcount = 0;
  stringflag = 0;
  for( ; ; )
  {
    if( !( stringflag ) )
    {
      wordsrc = scStrNextWord( src );
      if( !( wordsrc ) )
        break;
      if( wordsrc != src )
        dst[dstindex++] = ' ';
      src = wordsrc;
    }
    else
    {
      if( !( *src ) )
        break;
    }

    /* Inside a string, we must be very careful what we modify... */
    if( *src == '\"' )
    {
      if( !( stringflag ) )
        stringflag = 1;
      else if( src[-1] != '\\' )
        stringflag = 0;
    }

    wordlen = scStrWordLen( src );
    if( wordlen )
    {
      if( scStrCmpSeq( src, "include" ) ) 
      {
        linesrc = scStrNextLine( src );
        wordlen = linesrc - src;
        memcpy( &dst[dstindex], src, wordlen );
        src += wordlen;
        dstindex += wordlen;
        linelength = 0;
        src = linesrc;
        continue;
      }
      else if( scStrCmpSeq( src, "else{}" ) ) 
      {
        src += strlen( "else{}" );
        continue;
      }
      key = scNameKey( &keydict, src, wordlen );
      if( key )
      {
#ifdef SC_DEBUG_STDOUT
        fprintf( scfdlog, "Replace %s by %s\n", key->name, key->code );
#endif
        codelen = strlen( key->code );
        memcpy( &dst[dstindex], key->code, codelen );
        src += wordlen;
        dstindex += codelen;
        linelength += codelen;
        keymatchcount++;
      }
      else
      {
        memcpy( &dst[dstindex], src, wordlen );
        src += wordlen;
        dstindex += wordlen;
        linelength += wordlen;
      }
    }
    else
    {
      dst[dstindex++] = *src++;
      linelength++;
      if( !( stringflag ) && ( linelength >= linemaxlength ) && ( scSafeBreakChar[ (int)src[-1] ] ) )
      {
        dst[dstindex++] = '\r';
        dst[dstindex++] = '\n';
        linelength = 0;
      }
    }
  }
  end1:
  dst[dstindex] = 0;

  /*  End second pass */
  scKeyDictBuild( &keydict, mincodelen );
  fprintf( scfdlog, "  Key match Count : %d\n", keymatchcount );


  /* Reader headers to insert at top of script */
  header = 0;
  headersize = 0;
  if( ( fd = fopen( SC_FILE_HEADER, "rb" ) ) )
  {
    fseek( fd, 0, SEEK_END );
    headersize = ftell( fd );
    fseek( fd, 0, SEEK_SET );
    header = malloc( headersize+2 );
    if( header )
      fread( header, 1, headersize, fd );
    header[headersize+0] = '\r';
    header[headersize+1] = '\n';
    headersize += 2;
    fclose( fd );
  }


  /* Finally write the result */
  fprintf( scfdlog, "Writing output file %s\n", fileout );
  fd = fopen( fileout, "wb" );
  if( !( fd ) )
  {
    fprintf( scfdlog, "ERROR: Failed to write : %s\n", fileout );
    return 0;
  }
  if( header )
    fwrite( header, 1, headersize, fd );
  fwrite( buf0, 1, dstindex, fd );
  fprintf( fd, "\r\n// This Galaxy script was mangled by the Map obfuscation tool by Alexis Naveros\r\n" );
  fclose( fd );
  fprintf( scfdlog, "  Output Size : %d bytes\n", dstindex );


  /* Success, close log */
  fprintf( scfdlog, "Success!\n" );
  fclose( scfdlog );


  return 1;
}



